/*
	Internal graphics machine and buffer
*/

static char rcsid[]="$Id: buffer.c,v 1.2 1994/09/17 21:51:09 white Exp $";

#include "spring.h"
#include "graphics.h"

#include <stdio.h>
#include <string.h>

#ifdef DOS
char buffer[64000];
#endif

char *bufferdata=NULL;

static int xsize,ysize;

/* the actual drawing function to use */
static void (*buffer_drawfunc)(int,int,char);

static unsigned long buffer_drawmode=0L;
static unsigned long buffer_reflect=0L;

#define	REFLECT_MASK	3L

/* Additive (normal) drawing function */

void buffer_drawfuncadd(int x,int y,unsigned char c)
{
  /* maybe do a mod palette_size here? */

  bufferdata[x+xsize*y]+=c;

  if(buffer_reflect)
  {
	switch(buffer_reflect)
	  {
	  case 1L:		/* reflect in the y axis */
		bufferdata[xsize*y+(xsize-x-1)]+=c;
		break;
	  case 2L:		/* reflect in the x axis */
		bufferdata[x+xsize*(ysize-y-1)]+=c;
		break;
	  case 3L:		/* reflect in both axes */
		bufferdata[xsize*y+(xsize-x-1)]+=c;
		bufferdata[x+xsize*(ysize-y-1)]+=c;
		bufferdata[(xsize-x-1)+xsize*(ysize-y-1)]+=c;
		break;
	  }
  }
}		


/* Absolute drawing function (set by drawmode & 4) */

void buffer_drawfuncabs(int x,int y,unsigned char c)
{
  /* maybe do a mod palette_size here? */

  bufferdata[x+xsize*y]=c;

  if(buffer_reflect)
  {
	switch(buffer_reflect)
	  {
	  case 1L:		/* reflect in the y axis */
		bufferdata[xsize*y+(xsize-x-1)]=c;
		break;
	  case 2L:		/* reflect in the x axis */
		bufferdata[x+xsize*(ysize-y-1)]=c;
		break;
	  case 3L:		/* reflect in both axes */
		bufferdata[xsize*y+(xsize-x-1)]=c;
		bufferdata[x+xsize*(ysize-y-1)]=c;
		bufferdata[(xsize-x-1)+xsize*(ysize-y-1)]=c;
		break;
	  }
  }
}		

/* It's another level of indirection and another call,
   but it allows me to play around with drawing modes later.
*/

void buffer_point(int x,int y,unsigned char c)
{
  (*buffer_drawfunc)(x,y,c);
}

void buffer_setdrawmode(unsigned long dm)
{
  buffer_drawmode=dm;
  buffer_reflect=dm & REFLECT_MASK;
  buffer_drawfunc=
	(dm & 4L) ? buffer_drawfuncabs : buffer_drawfuncadd;
}

void buffer_init()
{
  xsize=graph_xsize();
  ysize=graph_ysize();
  buffer_setdrawmode(0L);

  if(!xsize)panic("can't initialise buffer before graphics");

  /* No, I know this is silly. My fault for not
	being able to afford a *real* DOS compiler. */
	
#ifdef DOS
  bufferdata=&buffer[0];
#else
  bufferdata=calloc(xsize,ysize);
#endif

  if(!bufferdata)panic("can't allocate buffer memory");
}

void buffer_rect(int x1,int y1,int x2,int y2,int c)
{
  int x,y;
  for(x=x1;x<x2;x++)
	{
	  graph_check();
	  for(y=y1;y<y2;y++)
		buffer_point(x,y,c);
	}
}
 
void buffer_clear()
{
	if(bufferdata)memset(bufferdata,'\0',xsize*ysize);
}
