#include <stdio.h>
#include <stdarg.h>
#include <string.h>                

#include "spring.h"
#include "compiler.h"
#include "nodetypes.h"
#include "opcodes.h"

/*
 *
 *
 */
/*
 * Code generator
 *
 */
/*
 *
 *
 */
static char rcsid[]="$Id: cg.c,v 1.4 1994/09/26 21:10:49 white Exp $";

static unsigned long curpos=0L;
extern unsigned long *codebuffer;
extern LIST *locals,*functions,*args;
LIST *badgotos,*labels;
extern LIST *addlist(LIST *,char *,int,...);
extern void deletelist(LIST *);
extern void deletetree(NODE *);
extern void optimise(void);
extern LIST *find(LIST *,char *);
void outputtree(NODE *t);

extern LIST *functab[];

extern short optim_on,show_out;

LIST *inter=NULL;	/* intermediate code */
LIST *intertail=NULL;

void panic(char *,...);

/* debugging stuff */

extern short debug;

char *show(unsigned long c)
{
  extern char *getlibcallname(int n);
  char buf[80];
  static char code[80];
  static char *opcodes[]=
  {
	"label","","","","",
	"jobless","dup","drop","minus","equals",
	"notequal","greater","less","gtequal","ltequal",
	"plus","div","mul","mod","and",
	"or","neg","not","getvar","getarg",
	"getelem","redundant","assign","assignarg",
	"assignelem","redundant","makeassoc","makearray",
	"self","inheritedcallbystring","freevec","makestr","freestr","null",
	"getargs","coordpair","arrowfromstring","arrowfromid","arrowfromobj",
	"arrowreturn","sleep"
  };

  sprintf(code,"%8lx  ",c);
  if(!(c & OPF_COMMAND))
	{
	  if(c&OPF_NEGATIVE) c|=0xc0000000L;
	  sprintf(buf,"Literal %ld",c);
	}
  else if(c & OPF_LIBCALL)
	{
	  sprintf(buf,"LIBRARY CALL %s",getlibcallname(c& 0x0ffffffL));
	}
  else if(c & OPF_FUNCCALL)
	{
	  sprintf(buf,"FUNCTION %d [%s]",c& 0x0ffffffL,
			  functab[c&0xffffffL]->str);
	}
  else if(c&OP_GOTO)
	{
	  unsigned long jp=c & 0xffff;
	  switch(c&0x0fff0000)
		{
		case OP_GOTO:strcpy(buf,"goto");break;
		case OP_GOTO_EQ:strcpy(buf,"goto/eq");break;
		case OP_GOTO_TRUE:strcpy(buf,"goto/true");break;
		case OP_GOTO_FALSE:strcpy(buf,"goto/false");break;
		default:sprintf(buf,"badop%lx",c);break;
		}
	}
  else
	strcpy(buf,opcodes[c&0x000000ff]);

  strcat(code,buf);	
  return(code);
}


/*
 *
 *
 */

#define	mklab(x)	_mklab((x),"_T")
#define	mkstlab(x)	_mklab((x),"_S")
#define	mkmlab(x)	_mklab((x),"_M")

static int nextlab=0,nextmlab=0;
static char *_mklab(int n,char *root)
{
  static char buf[20];

  sprintf(buf,"%s%d",root,n);
  return(buf);
}

void addinter(char *str,unsigned long n)	/*add to end */
{
  LIST *new;

  if(!inter)
  {
	inter=intertail=addlist(inter,str,3,n,0,0);
	/* kludge because int!=long under DOS */
	inter->nums[0]=n;
  }
  else
	{
	  new=malloc(sizeof(LIST));
	  if(!new)panic("no memory in intermediate stage");
	  strcpy(new->str,str);
	  new->nums[0]=n;
	  new->nums[1]=0;
	  new->nums[2]=0L;

	  new->next=NULL;
	  intertail->next=new;
	  intertail=new;
	}
}


void routput(unsigned long n)
{
  codebuffer[curpos]=n;
  if(curpos++==CODESIZE)
	panic("code too big!!");
}

void routputgoto(char *lab,unsigned long type)
{
  LIST *l;

  if(l=find(labels,lab))	/* previous label */
	{
	  routput(GOTO(type,l->nums[0]));
	}
  else
	{
	  /* this is a forward jump: output a null statement
		 and add to the list of Bad Gotos. Note that the goto type
		 is saved. */

	  badgotos=addlist(badgotos,lab,2,curpos,0);
	  routput(type);
	}
}

void routputlabel(char *lab)
{
  /* we never actually output a label, as such. First, we
	 add it to the label list, then we check the bad goto list
	 for any gotos which need to point here. */

  LIST *l;
  int pos;

  if(find(labels,lab))panic("Label defined twice - %s",lab);
  labels=addlist(labels,lab,2,curpos,0,0);

  for(l=badgotos;l;l=l->next)
	{
	  pos=l->nums[0];
	  if(!strcmp(lab,l->str))
		{
		  /* rebuild the jump from the saved goto type */
		  codebuffer[pos]=GOTO(codebuffer[pos],curpos);
		}
	}
}

void buildbuffer(void)
{
  LIST *l;
	

  for(l=inter;l;l=l->next)
	{
	  switch(l->nums[2])
		{
		case OP_LABEL:
		  routputlabel(l->str);
		  if(show_out){printf("   %s:\n",l->str);fflush(stdout);}
		  break;
		case OP_GOTO:
		case OP_GOTO_TRUE:
		case OP_GOTO_FALSE:
		case OP_GOTO_EQ:
		  routputgoto(l->str,l->nums[2]);
		  if(show_out)
			{
			  printf("        %s %s\n",show(GOTO(l->nums[2],0)),l->str);
			  fflush(stdout);
			}
		  break;
		case OP_JOBLESS:
		  if(show_out)printf("        OPTIMISED CODE %s(%d)\n",
							 l->str,l->nums[1]);
		  break;
		default:
		  if(show_out)printf("        %s\n",show(l->nums[0]));
		  routput(l->nums[0]);
		}
	}
  deletelist(inter);
  inter=NULL;
}

int flushbuffer(void)
{
  int i;

  if(optim_on)optimise();

  buildbuffer();

  deletelist(badgotos);badgotos=NULL;
  deletelist(labels);labels=NULL;
  return(i);
}

void output(unsigned long n)	/* add to intermediate code */
{
  addinter("instr",n);
}

void outputgoto(char *lab,unsigned long type)
{
  addinter(lab,OP_GOTOPLC);intertail->nums[2]=type;
}

void outputlabel(char *lab)
{
  addinter(lab,OP_LABELPLC);intertail->nums[2]=OP_LABEL;
}

/*
 * Do constant folding on a tree
 *
 */

void fold(NODE *t)
{
  static int nest=0;
  int i;
  if(!t)return;

  nest++;

  for(i=0;i<5;i++)fold(t->nodes[i]);	/* fold the leaves */
  if(t->type < 50) /* binops are down here */
	{
	  if((t->nodes[0]->type==NT_NUMBER) && (t->nodes[1]->type==NT_NUMBER))
		{
		  long a,b,c;
		  a=t->nodes[0]->value.num;
		  b=t->nodes[1]->value.num;

		  switch(t->type)
			{
			case NT_SUB:c=a-b;break;
			case NT_ADD:c=a+b;break;
			case NT_DIV:c=a/b;break;
			case NT_MUL:c=a*b;break;
			case NT_MOD:c=a%b;break;
			case NT_AND:c=a&&b;break;
			case NT_OR:c=a||b;break;
			case NT_EQUAL:c=a==b;break;
			case NT_NOTEQUAL:c=a!=b;break;
			case NT_LTEQUAL:c=a<=b;break;
			case NT_GTEQUAL:c=a>=b;break;
			case NT_GREATER:c=a>b;break;
			case NT_LESS:c=a<b;break;
			default:panic("bad node - %d",t->type);
			} /* switch */

		  t->type=NT_NUMBER;
		  deletetree(t->nodes[0]);t->nodes[0]=NULL;
		  deletetree(t->nodes[1]);t->nodes[1]=NULL;
		  t->value.num=c;

		} /* if both numbers */
	} /* if binop */
  else if(t->type < 100)
	{
	  fold(t->nodes[0]);
	  if(t->nodes[0]->type==NT_NUMBER)
		{
		  short a,c;
		  a=t->nodes[0]->value.num;

		  switch(t->type)
			{
			case NT_NEGATE:c=-a;break;
			case NT_NOT:c=!a;break;
			default:break;
			}
		  t->type=NT_NUMBER;
		  deletetree(t->nodes[0]);t->nodes[0]=NULL;
		  t->value.num=c;
		}
	}
  nest--;
}

void outputreturn(void)
{
  output(OP(OP_NULL));
}

void outputtree(NODE *t)
{
  static int nest=0;
  if(!t)return;

  nest++;

  fold(t);	/* fold any constants */

  switch(t->type)
	{
	case NT_STATEMENT:
	case NT_BLOCK:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  break;
	case NT_CASE:
	case NT_NULL:
	  break;
	case NT_SWITCH:
	  {
		NODE *n;int lab;

		/* stack the expression */
		outputtree(t->nodes[0]);

		/* go through the cases,
		   output the jumps and work out the labels for
		   each case */

		for(n=t->nodes[1];n;n=n->nodes[1])
		  {
			lab=nextlab++;

			/* If there is a next case, DUP the value for it. */
			if(n->nodes[1])
			  output(OP(OP_DUP));

			output(n->value.num & ~OPF_COMMAND);
			outputgoto(mklab(lab),OP_GOTO_EQ);
			/* change the case value to a label for later */
			n->value.num=lab;
		  }
		/* output the default case */
		outputtree(t->nodes[2]);

		/* generate jump to out of case */
		lab=nextlab++;
		outputgoto(mklab(lab),OP_GOTO);

		/* generate code for the cases */

		for(n=t->nodes[1];n;n=n->nodes[1])
		  {
			outputlabel(mklab(n->value.num)); /* made earlier */
			if(n->nodes[1]) /* if not the last case */
			  output(OP(OP_DROP));	/* generate a drop */
			outputtree(n->nodes[0]);	/* yer actual statement */
			outputgoto(mklab(lab),OP_GOTO); /* end of case */
		  }
		outputlabel(mklab(lab));	/* end of case label */
	  }
	  break;
	case NT_FOR:
	  {
		int lab1=nextlab++,lab2=nextlab++;
		outputtree(t->nodes[0]);	/* start statement */
		outputlabel(mklab(lab1));	/* loop start */
		outputtree(t->nodes[1]);	/* condition */
		outputgoto(mklab(lab2),OP_GOTO_FALSE); /* exit if false*/
		outputtree(t->nodes[3]);	/* body */
		outputtree(t->nodes[2]);	/* termination */
		output(OP(OP_DROP));		/* and a drop */
		outputgoto(mklab(lab1),OP_GOTO);	/* goto start */
		outputlabel(mklab(lab2));	/* exit point */
	  }
	  break;
	case NT_GOTO:
	  outputgoto(t->value.str,OP_GOTO);
	  break;
	case NT_LABEL:
	  outputlabel(t->value.str);
	  break;
	case NT_WHILE:
	  {
/*
 * This is done by generating an if statement, the true part of
 * which has a goto tacked on the end to jump back to the test.
 * while(TRUE) is optimised.
 */
		int lab1=nextlab++,lab2=nextlab++;

		outputlabel(mklab(lab1));
		fold(t->nodes[0]);
		if(t->nodes[0]->type==NT_NUMBER) /* is this "while(true)"? */
		  {
			if(t->nodes[0]->value.num)
			  {
				/* output a plain loop */
				outputtree(t->nodes[1]);
				outputgoto(mklab(lab1),OP_GOTO);
			  }
			else panic("while statement never true");
		  }
		else
		  {
			outputtree(t->nodes[0]);
			outputgoto(mklab(lab2),OP_GOTO_FALSE);
			outputtree(t->nodes[1]);
			outputgoto(mklab(lab1),OP_GOTO);
			outputlabel(mklab(lab2));
		  }
	  }
	  break;
	case NT_IF:
	  {
		int lab1=nextlab++,lab2=nextlab++;

		/* have I folded the condition out of existence ? */

		if(t->nodes[0]->type==NT_NUMBER)
		  panic(t->nodes[0]->value.num ?
				"if statement always true":
				"if statement always false");

		outputtree(t->nodes[0]); /* expression */
		outputgoto(mklab(lab1),OP_GOTO_FALSE);
		outputtree(t->nodes[1]);
		if(t->nodes[2])outputgoto(mklab(lab2),OP_GOTO);
		outputlabel(mklab(lab1));
		outputtree(t->nodes[2]);
		if(t->nodes[2])outputlabel(mklab(lab2));
	  }
	  break;
	case NT_FUNCCALL:
	case NT_LIBCALL:
	  {
		int nargs=0;	/* keep an argument count */
		NODE *n;
		for(n=t->nodes[0];n;n=n->nodes[4])
		  {
			outputtree(n);nargs++;
		  }
			  
		/* If this is a varargs function, stack the number
		   of arguments */

		if(t->value.str[0] & FUNCB_VARARGS)
		  output(nargs);

		switch(t->type)
		  {
		  case NT_LIBCALL:
			output(LIBCALL(t->value.num));break;
		  case NT_FUNCCALL:
			output(FUNCCALL(t->value.num));break;
		  default:panic("oops!");
		  }
		/* returns a value which is unused */
		if((t->value.str[0] & FUNCB_RETURNS) &&
		   (t->value.str[0] & FUNCB_ISSTMT))output(OP(OP_DROP));
	  }
	  break;
	case NT_SUB:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_MINUS));
	  break;
	case NT_EQUAL:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_EQUALS));
	  break;
	case NT_NOTEQUAL:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_NOTEQUAL));
	  break;
	case NT_GREATER:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_GREATER));
	  break;
	case NT_LESS:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_LESS));
	  break;
	case NT_GTEQUAL:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_GTEQUAL));
	  break;
	case NT_LTEQUAL:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_LTEQUAL));
	  break;
	case NT_ADD:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_PLUS));
	  break;
	case NT_DIV:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_DIV));
	  break;
	case NT_MUL:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_MUL));
	  break;
	case NT_MOD:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_MOD));
	  break;
	case NT_AND:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_AND));
	  break;
	case NT_OR:
	  outputtree(t->nodes[0]);
	  outputtree(t->nodes[1]);
	  output(OP(OP_OR));
	  break;
	case NT_NEGATE:
	  outputtree(t->nodes[0]);
	  output(OP(OP_NEG));
	  break;
	case NT_NOT:
	  outputtree(t->nodes[0]);
	  output(OP(OP_NOT));
	  break;
	case NT_NUMBER:
	  /* make sure the command bit is clear */
	  output(t->value.num & ~OPF_COMMAND);
	  outputtree(t->nodes[1]);
	  break;
	case NT_GETVAR:
	  output(t->value.num);
	  output(OP(OP_GETVAR));
	  outputtree(t->nodes[1]);	/* list linkage */
	  break;
	case NT_GETARG:
	  output(t->value.num);
	  output(OP(OP_GETARG));
	  outputtree(t->nodes[1]);	/* list linkage */
	  break;
	case NT_ASSIGN:
	  outputtree(t->nodes[0]);
	  output(t->value.num);
	  output(OP(OP_ASSIGN));
	  if(t->value.str[0])
		output(OP(OP_DROP));
	  break;
	case NT_ASSIGNARG:
	  outputtree(t->nodes[0]);
	  output(t->value.num);
	  output(OP(OP_ASSIGNARG));
	  if(t->value.str[0])
		output(OP(OP_DROP));
	  break;
	case NT_RETURN:	/* just stack and end */
	  outputtree(t->nodes[0]);
	  output(OP(OP_NULL));
	  break;

	default:
	  panic("bad node - %d",t->type);
	}
  nest--;
}

void yyraise(char *a,...)
{
  va_list argptr;
  static char buf[128];
  extern short lines;
  extern unsigned long pc,last_inst;
  extern void graph_shutdown();

  va_start(argptr,a);
  vsprintf(buf,a,argptr);

  graph_shutdown();
  if(lines>=0)
	printf("Heavy! Error at line %d: %s\n",lines,buf);
  else
  {
  	printf("Heavy! Error at instruction %ld (%s):\n%s\n",
  		pc-1,show(last_inst),buf);
  	interp_backtrace();
  }
  exit(0);
}

void yyerror(char *s)
{
  panic(s);
}

void panic(char *a,...)
{
  va_list argptr;
  char buf[80];

  va_start(argptr,a);
  vsprintf(buf,a,argptr);
  yyraise(buf);
}

unsigned long getcompilerpos(void)
{
  return curpos;
}

void zerocompilerpos(void)
{
  curpos=0L;
}
