/*
	Interpreter for virtual machine
*/

static char rcsid[]="$Id: inter.c,v 1.3 1994/09/24 23:43:19 white Exp $";

#include <stdio.h>
#include "spring.h"
#include "inter.h"
#include "graphics.h"
#include "compiler.h"
#include "opcodes.h"

#define	NOT_RUNNING	0xffffffff

extern LIST *functions;
extern unsigned long *codebuffer;
extern LIST *find(LIST *,char *);
extern LIST *functab[],*imagefunctab[];
extern int ifuncct;
extern char *show(long);

/* the return stack */
static struct
{
  int locals[MAXVARS];
  int args[MAXVARS];
  unsigned long pc;
} rstack[STACKSIZE];
static long roffset=0;


/* the execution stack */
long estack[STACKSIZE];
long eoffset=0;

/* the context variables */
/* The context starts as NOT_RUNNING */
unsigned long pc=NOT_RUNNING;
unsigned long last_inst=0L;

static int locals[MAXVARS],args[MAXVARS];



/* stack management routines */


/* Pop the context stack */
static void rPop(void)
{
#if	STACK_CHECK
  if(!roffset)panic("return stack underflow");
#endif
  roffset--;
  memcpy(locals,rstack[roffset].locals,sizeof(locals));
  memcpy(args,rstack[roffset].args,sizeof(args));
  pc=rstack[roffset].pc;
}

/* Push the context stack */
static void rPush(void)
{
#if	STACK_CHECK
  if(roffset==STACKSIZE)panic("return stack overflow");
#endif
  memcpy(rstack[roffset].locals,locals,sizeof(locals));
  memcpy(rstack[roffset].args,args,sizeof(args));
  rstack[roffset++].pc=pc;
}

/* only compile these if we're using not using macros */

#if !STACK_MACROS

/* Pop the execution stack */
long Pop()
{
#if	STACK_CHECK
  if(!eoffset)panic("stack underflow - too few args to function?");
#endif
  return(estack[--eoffset]);
}

/* what's atop the stack? */
long Peek(void)
{
#if	STACK_CHECK
  if(!eoffset)panic("stack underflow in peek");
#endif
  return(estack[eoffset-1]);
}

#endif	/* STACK_MACROS */
/* Push the execution stack */
void Push(long val)
{
#if	STACK_CHECK
  if(eoffset==STACKSIZE)panic("stack overflow");
#endif
  estack[eoffset++]=val;
}


/* initialise the interpreter */
void interp_init()
{
  LIST *f;

  /* Stack the initial "not running" context */
  rPush();

  /* find the main function */
  
  f=find(functions,"main");
  if(!f)panic("er.. I'd quite like a 'main' function.");

  /* and this is where we start */
  pc=f->nums[2];
}

/* run a single instruction */
void interp_run()
{
  extern void (*LibFuncs[])();
  unsigned long inst;
  extern int holding;
  long a,b;

  for(;;)
	{
	  graph_check();	/* check palette timing */
	  /* and spin while holding */
	  while(holding)graph_check();
	  
	  last_inst=inst=codebuffer[pc++];

	  dprintf("        RUNNING %ld:%s[%lx] (stack:%d - top %ld)\n",pc-1L,
				show(inst),inst,eoffset,eoffset?Peek():0L);
	
	
	  if(inst & OPF_COMMAND)				/* is it a command? */
		{
		  inst &= ~OPF_COMMAND;
		  if(inst & OPF_LIBCALL)		
			{
			  (*LibFuncs[inst&~OPF_LIBCALL])();
			}
		  else if(inst & OPF_FUNCCALL)
			{
			  rPush();
			  dprintf("Executing function number %d\n",inst);
			  inst &= ~OPF_FUNCCALL;
			  pc=functab[inst]->nums[2];
			}
		  else if(inst>= (1L<<16L))	/*is it a goto? */
			{
			  unsigned long dest=inst & ((1L<<16L)-1);	/* destination? */
			  switch(inst-dest) /* what sort? */
				{
				case OP_GOTO:
				  pc=dest;break;
				case OP_GOTO_TRUE:
				  if(Pop())pc=dest;break;
				case OP_GOTO_FALSE:
				  if(!Pop())pc=dest;break;
				case OP_GOTO_EQ:
				  if(Pop()==Pop())pc=dest;break;
				default:
				  panic("bad instruction/goto - %ld:%lx",pc-1L,
					inst);
				}
			}
		  else
			{
		  
			  /****Sat May  7 18:37:45 1994*******************************************
			   *
			   * The instruction we have grabbed is an opcode. Execute the
			   * corresponding operation.
			   ***********************************************************************/
	
			  switch(inst)
				{
				  /*	Housekeeping operations */
	 						
				case OP_DUP:
				  Push(Peek());break;
				case OP_DROP:
				  Pop();break;

				  /*	Arithmetic operations */
	 						
				case OP_MINUS:
				  b=Pop();a=Pop();
				  Push(a-b);break;
				case OP_EQUALS:
				  b=Pop();a=Pop();
				  Push(a==b);break;
				case OP_NOTEQUAL:
				  b=Pop();a=Pop();
				  Push(a!=b);break;
				case OP_GREATER:
				  b=Pop();a=Pop();
				  Push(a>b);break;
				case OP_LTEQUAL:
				  b=Pop();a=Pop();
				  Push(a<=b);break;
				case OP_LESS:
				  b=Pop();a=Pop();
				  Push(a<b);break;
				case OP_GTEQUAL:
				  b=Pop();a=Pop();
				  Push(a>=b);break;
				case OP_PLUS:
				  b=Pop();a=Pop();
				  Push(a+b);break;
				case OP_DIV:
				  b=Pop();a=Pop();
				  if(!b)panic("division by zero");
				  Push(a/b);break;
				case OP_MUL:
				  b=Pop();a=Pop();
				  Push(a*b);break;
				case OP_MOD:
				  b=Pop();a=Pop();
				  if(!b)panic("modulo by zero");
				  Push(a%b);break;
				case OP_AND:
				  b=Pop();a=Pop();
				  Push(a&&b);break;
				case OP_OR:
				  b=Pop();a=Pop();
				  Push(a||b);break;
				case OP_NEG:
				  Push(-Pop());break;
				case OP_NOT:
				  Push(!Pop());break;
	
				  /*	variable, array and string management */
	
				case OP_GETVAR:
				  Push(locals[Pop()]);break;
				case OP_GETARG:
				  Push(args[Pop()]);break;
				case OP_ASSIGN:
				  a=Pop();
				  b=Peek();
				  locals[a]=b;
				  break;
				case OP_ASSIGNARG:
				  a=Pop();
				  b=Peek();
				  args[a]=b;
				  break;
				case OP_NULL:
				  /* effectively a return. Called NULL for
					 hysterical raisins.*/
	
				  rPop();
				  if(interp_isfinished())return;

				  break;
				case OP_GETARGS:
				  a=Pop();
				  while(a>0L)
					{
					  args[--a]=Pop();
					}
				  break;
				default:
				  panic("bad instruction: %ld:%lx",pc-1L,inst);
				} /* switch */
			} /* else */
		} /* if inst&OPF_COMMAND */
	  else
		{
		  if(inst & OPF_NEGATIVE)
			inst |= 0x80000000L;
		  Push(inst);				
		}
	}
}

/* returns true if the interpreter has completed */
short interp_isfinished()
{
  return(pc==NOT_RUNNING);
}

/* start the interpreter off on a new function */
void interp_function(short n)
{
  rPush();
  pc=functab[n]->nums[2];
}

static LIST **images=NULL;
static LIST *currentimage=NULL;

/*
	Shuffle the interpreter functions
*/

#define	SHUFFLETIMES	10

static void interp_shuffle()
{
	int i,a,b;
	LIST *p;
	
	/* if no internal image list exists, build one and copy
	the image functions into it, unshuffled. */
	if(!images)
	{
		images=(LIST **)malloc(sizeof(LIST **)*ifuncct);
		for(i=0;i<ifuncct;i++)images[i]=imagefunctab[i];
	}

	/* Now shuffle - swap items over, at random, SHUFFLETIMES times */

	for(i=0;i<SHUFFLETIMES;i++)
	{
		a=rand()%ifuncct;
		b=rand()%ifuncct;
		if(a==b)continue;
		p=images[a];
		images[a]=images[b];
		images[b]=p;
	}
}
		


void interp_choose()
{
  static int inum=0;
  
  if(!ifuncct)panic("No fair! No image functions to choose from!");

  if(!inum)interp_shuffle();
  
  rPush();
  currentimage=images[inum++];
  pc=currentimage->nums[2];
  if(inum==ifuncct)inum=0;
}

void interp_backtrace()
{
	printf("Stack trace:\n");
	while(roffset>0)
	{
		printf("     %10ld  %s\n",
			pc-1,show(codebuffer[pc-1]));
		rPop();
	}
}

/* get the name of the current image function */
char *interp_name()
{
  static char *err="<<no current function>>";
  if(currentimage)
	return(currentimage->str);
  else
	return(err);
}

/* panic which returns dummy int. */
int ipanic(char *a)
{
  panic(a);
}
