%{
#include <stdio.h>
#include <stdlib.h>
#include "compiler.h"
#include "spring.h"

static char rcsid[]="$Id: lexer.l,v 1.3 1994/09/22 18:32:54 white Exp $";

#ifdef	DOS
#ifdef	__TURBOC__
#define	__STDC__	1
#endif
#include "y_tab.h"
#else
#include "parser.tab.h"
#endif
extern void panic(char *,...);
extern LIST *addlist(LIST *,char *,int,...);
extern void deletelist(LIST *);
extern LIST *find(LIST *,char *);
void lex_end();

int lines=0;

#undef YY_INPUT
#define	YY_INPUT(buf,result,max_size)\
{ \
int c;\
if(!yyin)panic("no input file");\
c=getc(yyin);\
if(c==10)lines++;\
if(debug==2){putchar(c);if(c==10)printf("+%d ",lines);}\
result=(c==EOF)?YY_NULL:(buf[0]=c,1);\
}

#undef	yywrap()
#define	yywrap()	1

extern short debug;

void mknum(void);
void mkname(void);
void mkcharcon(void);
void mktext(void);

%}

comment	"//".*
delimiter	[\t\n\r ]
whitespace	{delimiter}+
uc_letter	[A-Z]
lc_letter	[a-z]
letter	{uc_letter}|{lc_letter}
ascii_char	[^\"]
escaped_char	\\{lc_letter}|\\\"
digit	[0-9]
idsymb	{letter}|{digit}|_
initidsymb	{letter}|_
id	{initidsymb}{idsymb}*
number	{digit}+
text	\"({ascii_char}|{escaped_char})*\"
charcon	\'({ascii_char}|{escaped_char})\'

%%

{comment}	{ }
{whitespace}	{ }

\|\|		{return(OR);}
&&		{return(AND);}
void	{return(VOID);}
if		{return(IF);}
else		{return(ELSE);}
goto		{return(GOTOTOK);}
while		{return(WHILE);}
\=\=		{return(EQUAL);}
\<=			{return(LTEQUAL);}
\>=			{return(GTEQUAL);}
!=|\<\>			{return(NOTEQUAL);}
"return"		{return(RETURN);}
for			{return(FOR);}
default		{return(DEFAULT);}
case		{return(CASE);}
switch		{return(SWITCH);}
local		{return(LOCAL);}
auto		{return(LOCAL);}
image		{return(IMAGE);}
!			{return(NOT);}

{id}		{ mkname(); return(ID); }
{text}		{ mktext(); return(STRING); }
{number}	{ mknum(); return(NUMBER); }
{charcon}	{ mkcharcon(); return(NUMBER); }
<<EOF>>		{ return(ENDOFFILE);}
.			{return(*yytext);}

%%

/*
 * Support routines
 *
 */

void lex_end()
{
  yy_init_buffer(yy_current_buffer,NULL);
}

void mkname(void)
{
  strcpy(yylval.str,yytext);

}

static void process_text(char *txt)
{
  char *tmp,*p,*q;

  tmp=malloc(strlen(txt));	/* get space for the string */

  for(q=tmp,p=txt;*p;p++)
	{
	  if(*p=='\\')	/* it's an escaped char! */
		{
		  switch(*++p)
			{
			case 'n':*q++='\n';break;
			case 't':*q++='\t';break;
			case 'v':*q++='\v';break;
			case 'b':*q++='\b';break;
			case 'r':*q++='\r';break;
			case 'f':*q++='\f';break;
			case 'a':*q++='\a';break;
			default: *q++=*p;break;
			}
		}
	  else
		{
		  *q++=*p;
		}
	}
  *q='\0';
  strcpy(txt,tmp);	/* and copy the new version back */
  free(tmp);
}

/*
	This routine will add a text to the StringManager lists, and return
	the ID
*/

void mktext(void)
{
  char *tmp,*p,*q;
/*
  extern unsigned long NewString(char *str);
  
  yytext[strlen(yytext)-1]='\0';  

  process_text(yytext+1);
  yylval.num=(int)NewString(yytext+1);
*/
panic("no strings, sorry!");
}

void mknum(void)
{

	yylval.num=atol(yytext);
}

void mkcharcon(void)
{
  int i;
  unsigned long total=0L,mult=1L;

  /* allows 'extended' word-sized character constants */

  yytext[strlen(yytext)-1]='\0';

  process_text(yytext+1);
  for(i=strlen(yytext)-1;i>0;i--)
	{
	  total+=yytext[i]*mult;
	  mult *= 256L;
	}
  yylval.num=total;
}
