static char rcsid[]="$Id: main.c,v 1.6 1994/09/28 21:44:00 white Exp white $";

#define	VERSION	"0.2"

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include "spring.h"

#include "graphics.h"
#include "buffer.h"
#include "palette.h"
#include "math.h"
#include "inter.h"
#include "waves.h"

#define	DEFAULTFILE "default.sfe"

extern FILE *yyin;

void title()
{
	printf("\nThe Springfield Effect  - version %s, (C) 1994 Jim Finnis\n",
		VERSION);

	printf("(C) 1994 Jim Finnis (white@elf.dircon.co.uk)\n\n\n");
	printf("    This program is free software; you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation; either version 2 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program; if not, write to the Free Software\n\
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n");
	
	printf("You may have to wait some time for the first image...\n");
	sleep(2);
}

short debug=0,show_out=0,optim_on=1,show_time=0;
unsigned long *codebuffer=NULL;
char **argvec;
extern int yydebug;

void usage()
{
  fprintf(stderr,"The Springfield Effect: Usage\n");
  fprintf(stderr,"spring filename [-d] [-s] [-x<num>] [-y<num>] [-t<num>]\n");
  fprintf(stderr,"-d    : turn on debugging (i.e. start with trace(1))\n");
  fprintf(stderr,"-r    : specify resolution of window\n");
  fprintf(stderr,"-t    : slow down interrupts by a given factor. Handy on slow PCs\n");
  fprintf(stderr,"-s    : show VM instructions on stdout\n");
  fprintf(stderr,"-T    : print time taken for each image to stdout\n");
  exit(1);
}

/*
		Open a file, looking in . first, followed by the
		library directory
*/

FILE *pathfopen(char *n,char *mode)
{
	char buf[256];
	
	FILE *a;
	if(!(a=fopen(n,mode)))
	{
		if(!isalpha(*n))return(NULL);
		strcpy(buf,FILEPATH);
		strcat(buf,n);

		if(!(a=fopen(buf,mode)))
		{
			return(NULL);
		}
	}
	return(a);
}

void main(int argc,char **argv)
{
  extern int yyparse();
  extern int lines;
  char *fn;
  unsigned long t,t2;
  short scale=-1;	/* -1 means use defaults in 'graphics.c' */
  short timing=1;

  lines=-1;
 
  title();
  srand(time(&t));

  yydebug=0;
  fn=DEFAULTFILE;

  /* I don't hold with this newfangled 'getopt' stuff. */

  while(--argc>0)
	{
	  if(*argv[argc]=='-')
		{
		  switch(argv[argc][1])
			{
			case 'd':debug=1;break;
			case 'D':debug=2;break;
			case 's':show_out=1;break;
			case 'r':scale=atoi(argv[argc]+2);break;
			case 't':timing=atoi(argv[argc]+2);break;
			case 'T':show_time=1;break;
			default: usage();
			}
		}
	  else fn=argv[argc];
	}

  graph_setsize(scale);
  graph_settiming(timing);

  if(!(codebuffer=malloc(CODESIZE*sizeof(long))))
	panic("This is silly. Can't even allocate a codebuffer.");

  if(!(yyin=pathfopen(fn,"r")))panic("can't open file - %s",fn);

  yyparse();

  lines=-1;
  fclose(yyin);
  
  argvec=argv;		/* needed by things like X */

  /* Here we go.... */

  graph_init();		/* initialise machine-dependent graphics */
  palette_init();	/* initialise palette */
  buffer_init();	/* intialise the drawing buffer */
  math_init();		/* initialise math tables */
  wave_init();		/* initialise wave manager */
  interp_init();	/* initialise interpreter */

  printf("Scale: %d, (%dx%d)\n",graph_getscale(),graph_xsize(),
		 graph_ysize());

  /* set the initial grey palette */

  graph_palette(palette_get());

  /* Having initialised everything, with a quick prayer to Eris,
  we send it on its merry way */

  for(;;)
	{
	  time(&t);
	  /* get the picture */
	  interp_run();
	  time(&t2);
	  if(show_time)printf("Time for image %s = %ld\n",interp_name(),
						  t2-t);
	  graph_show();
	  interp_choose();

	  /* fade up */
	  palette_startfadein(1);
	}

  graph_shutdown();	/* close the graphics */
  printf("normal exit!\n");  
}
