#include <math.h>
#include <stdio.h>

#include "spring.h"
#include "math.h"
#include "graphics.h"

static char rcsid[]="$Id: math.c,v 1.2 1994/09/17 21:51:09 white Exp $";


/*
	Math routines and lookup table management
*/
      
#ifndef M_PI
#define M_PI	3.1415927
#endif

#define	DISTSIZ	200

short sinTab[4096];		/*sine table*/
short wSinTab[4096];		/*0-32767 sintable, for waves*/

#if __TURBOC__
short huge distTab[DISTSIZ][DISTSIZ];
#else
short distTab[DISTSIZ][DISTSIZ];/* distance table */
#endif

void math_init()
{
	long i,j;
	float f,angle_rads,pi2;

	pi2=2.0*M_PI;

	for(i=0;i<4096;i++)
	{
		f=(int)i;
		angle_rads=f*pi2/4096.0;

		sinTab[i]=(int)(sin(angle_rads)*32767.0);
		wSinTab[i]=sinTab[i]/2+16383;
	}

	/*
		Work out the distance table for the ripple and blur functions
	*/

	for(i=0;i<DISTSIZ;i++)
	{
		for(j=0;j<DISTSIZ;j++)
		{
			f=(float)(i*i+j*j);
			distTab[i][j]=(short)sqrt(f);
		}
	}
}

/*
	This returns the sin of a value, with units of
	0-4095 == 0-360deg.
*/
int mSin(int a)
{
  while(a<0)a+=4096;
  return(sinTab[a%4096]);
}

/*
	See above for units
*/
int mCos(int a)
{
  while(a<0)a+=4096;
  return(sinTab[(a+1024)%4096]);
}

/*
	returns the distance between two points
*/

int math_dist(int x1,int y1,int x2,int y2)
{
	long dx,dy,mul=0L;

	dx=(long)x1-(long)x2;if(dx<0L)dx= -dx;
	dy=(long)y1-(long)y2;if(dy<0L)dy= -dy;

	/* scale down if necessary */
	while(dx>=DISTSIZ || dy>=DISTSIZ)
	{
		dx/=2;dy/=2;
		mul++;
	}
	return(distTab[dx][dy]<<mul);
}
