#include <stdio.h>
#include <stdarg.h>

#include "compiler.h"
#include "opcodes.h"

static char rcsid[]="$Id: optimise.c,v 1.2 1994/09/17 21:51:09 white Exp $";

/*
 *
 *
 */
/*
 * Noddy Optimiser
 *
 */
/*
 *
 *
 */

extern LIST *inter;	/* intermediate code list */
void yyraise(char *,...);
extern short debug;

static LIST *findlabel(char *lab)
{
	LIST *l;
	for(l=inter;l;l=l->next)
	{
		if(l->nums[2]==OP_LABEL && !strcmp(lab,l->str))return(l);
	}
	yyraise("Label not found - %s\n",lab);
}

static LIST *getnext(LIST *l)
{
	l=l->next;
	while(l)
	{
		if(l->nums[2]!=OP_JOBLESS && l->nums[2]!=OP_GOTOPLC &&
		   l->nums[2]!=OP_LABELPLC)return(l);
		l=l->next;
	}
	return(NULL);
}

static void removeop(LIST *l,int code)
{
		if(debug)printf("removing code..\n");
		l->nums[2]=OP_JOBLESS;
		l->nums[1]=code;
}

/*
 * Optimise NOTs in goto conditions - i.e. if the condition is NOTed,
 * remove the NOT and negate the condition.
 *
 */

static void optimise_goto_conditions()
{
	LIST *l,*t;

	for(l=inter;l;l=getnext(l))
	{
		if(l->nums[0]==OP(OP_NOT))
		{
				if(t=getnext(l))
				{
					if(t->nums[2]==OP_GOTO_TRUE)
					{
						if(debug)printf("optimising not true goto %s\n",
							l->str);
						removeop(l,1);
						t->nums[2]=OP_GOTO_FALSE;
						strcpy(l->str,t->str);
					}
					else if(t->nums[2]==OP_GOTO_FALSE)
					{
						if(debug)printf("optimising not false goto %s\n",
							l->str);
						removeop(l,1);
						t->nums[2]=OP_GOTO_TRUE;
						strcpy(l->str,t->str);
					}
				}
		}
		if(l->nums[0]==OP(OP_EQUALS))
		{
				if(t=getnext(l))
				{
					if(t->nums[2]==OP_GOTO_TRUE)
					{
						if(debug)printf("optimising goto-equals %s\n",
							l->str);
						removeop(l,2);
						t->nums[2]=OP_GOTO_EQ;
						strcpy(l->str,t->str);
					}
				}
		}
	}
}

/*
 * Look for gotos to unconditional gotos, and change them to jumps
 * to the target of the second goto. Returns 1 if skipping was done.
 *
 */

static int skip_double_gotos()
{
	LIST *l,*t;
	int done=0;

	for(l=inter;l;l=getnext(l))
	{
		switch(l->nums[2])
		{
			case OP_GOTO:
			case OP_GOTO_FALSE:
			case OP_GOTO_TRUE:
			case OP_GOTO_EQ:
				t=findlabel(l->str);	/* find the destination */
				if(t=getnext(t))	/* after the label */
				{
					if(t->nums[2]==OP_GOTO)	/* another goto! */
					{
						if(debug)printf("skipping goto %s -> %s\n",
							l->str,t->str);
						strcpy(l->str,t->str);
						done=1;
					}
					if(t->nums[0]==OP(OP_NULL) && /*a retrn */
						l->nums[2] == OP_GOTO)

					{
						if(debug)printf("goto null at %s\n",
							l->str);
						l->nums[2]=0L;
						l->nums[0]=OP(OP_NULL);
						done=1;
					}
				}
		}
	}
	return(done);
}

/*
 * Remove gotos whose labels follow them immediately
 *
 */

static void remove_redundant_gotos()
{
	LIST *l,*t;

	for(l=inter;l;l=getnext(l))
	{
		switch(l->nums[2])
		{
			case OP_GOTO:
			case OP_GOTO_EQ:
			case OP_GOTO_FALSE:
			case OP_GOTO_TRUE:
			{
				if(t=getnext(l))
				{
					if(t->nums[2] == OP_LABEL && !strcmp(l->str,t->str))
					{
						if(debug)printf("removing redundant goto to %s\n",
							t->str);
						removeop(l,3);
					}
				}
			}
		}
	}
}

/*
 * Remove labels with no gotos to them
 *
 */

static void remove_unreffed_labels()
{
	LIST *l,*t;
	int found;

	for(l=inter;l;l=getnext(l))
	{
		if(l->nums[2]==OP_LABEL)
		{
			found=0;
			for(t=inter;t;t=getnext(t))
			{
				switch(t->nums[2])
				{
					case OP_GOTO:
					case OP_GOTO_FALSE:
					case OP_GOTO_TRUE:
					case OP_GOTO_EQ:
						if(!strcmp(l->str,t->str))found=1;break;
				}
				if(found)break;
			}
			if(!found)
			{
				if(debug)printf("removing redundant label %s\n",l->str);
				removeop(l,4);
			}
		}
	}
}

/*
 * Look for dead code,
 * i.e. unlabelled code following an unconditional branch
 */

static void remove_dead_code()
{
	LIST *l;
	int del=0;

	for(l=inter;l;l=getnext(l))
	{
		if(l->nums[2] == OP_LABEL)del=0;
		else if(!del &&
			((l->nums[2] == OP_GOTO)||(l->nums[0]==OP(OP_NULL))))
				del=1;
		else if(del){
				removeop(l,5);
			if(debug)printf("dead code removed\n");
		}
	}
}

/*
 * replace all references to label 1 with label 2
 *
 */

static void replace_label(char *lab1,char *lab2)
{
	LIST *l;
	if(debug)printf("replacing %s with %s\n",lab1,lab2);
	for(l=inter;l;l=getnext(l))
	{
		switch(l->nums[2])
		{
			case OP_GOTO:
			case OP_GOTO_FALSE:
			case OP_GOTO_TRUE:
			case OP_GOTO_EQ:
				if(!strcmp(lab1,l->str))strcpy(l->str,lab2);
				break;
			default:break;
		}
	}
}


/*
 * Search for consecutive labels, and replace the second one with
 * the first.
 *
 */

void remove_consecutive_labels(void)
{
	LIST *l,*t;
	for(l=inter;l;l=getnext(l))
	{
		if(l->nums[2]==OP_LABEL)
		{
			if(t=getnext(l))
			{
				if(t->nums[2]==OP_LABEL)
				{
					removeop(t,6);
					replace_label(t->str,l->str);
				}
			}
		}
	}
	remove_unreffed_labels();
}

/* remove goto-cond l1/goto l2/l1: */

void remove_goto_goto(void)
{
	LIST *a,*b,*c;
	for(a=inter;a;a=getnext(a))
	{
		if(!(b=getnext(a)))break;
		if(!(c=getnext(b)))break;

		if(a->nums[2]!=OP_GOTO_FALSE && a->nums[2]!=OP_GOTO_TRUE)
			break;
		if(b->nums[2] != OP_GOTO)break;	/* 2nd - uncond goto */

		/* 3rd is a label that the 1st jumps to */
		if(c->nums[2]==OP_LABEL && !strcmp(c->str,a->str))
		{
			/* Do it! */
if(debug)
printf("replacing\n  goto/cond %s\n  goto %s\n%s:\nwith\n  goto/ncond %s\n",
	a->str,b->str,c->str,b->str);

			strcpy(a->str,b->str);	/* 2nd goto label into 1st */
			switch(a->nums[2])
			{
				case OP_GOTO_TRUE:
					a->nums[2]=OP_GOTO_FALSE;break;
				case OP_GOTO_FALSE:
					a->nums[2]=OP_GOTO_TRUE;break;
				default:break;
			}
			removeop(b,7);
			removeop(c,8);
		}
	}
}


void optimise(void)
{
	short i;

	if(debug)printf("\nOptimising code block\n");

	for(i=0;i<3;i++)
	{
		optimise_goto_conditions();
		remove_redundant_gotos();
		while(skip_double_gotos());
		remove_redundant_gotos();
		while(skip_double_gotos());
		remove_unreffed_labels();
		remove_dead_code();
		remove_redundant_gotos();
		remove_consecutive_labels();
		remove_redundant_gotos();
		remove_goto_goto();
		remove_dead_code();
	}
}
