/*
	Colour palette manipulation
*/

static char rcsid[]="$Id: palette.c,v 1.3 1994/09/22 16:50:49 white Exp $";

#include "palette.h"
#include "graphics.h"
#include "spring.h"
#include "inter.h"
#include <stdio.h>

static int size;
static short isinit=0;

/* 256 colours max, and a power of 2!
   Although colours are 24bit, this is 
   an array of short, not char, to allow
   for the method used in palette_run()
*/

static unsigned short pal[256][3];

short palette_isinit()
{
  return isinit;
}

void palette_init()
{
  int i,s;
  size=graph_psize();
  if(!size)panic("graphics not initialised in palette_init()");

  /* create the initial black palette */

  for(i=0;i<size;i++)
	{
	  pal[i][0]=0;
	  pal[i][1]=0;
	  pal[i][2]=0;
	}
  isinit=1;
}

short *palette_get()
{
  return (short *)pal;
}


/* Internal palettes are of variable size. */

struct palette
{
  unsigned char *r;
  unsigned char *g;
  unsigned char *b;
  int speed,size,ct;
};

static struct palette pals[MAXPALS];
static int pct=0;
static int fade=0,fadespeed=0;

unsigned long fadered,fadegreen,fadeblue;

long palette_new(int size)
{
  if(pct==MAXPALS)panic("too many palettes");
  pals[pct].size=size;
  pals[pct].speed=0;
  pals[pct].ct=0;
  if(!(pals[pct].r=calloc(size,sizeof(unsigned char))))panic("out of memory for palette");
  if(!(pals[pct].g=calloc(size,sizeof(unsigned char))))panic("out of memory for palette");
  if(!(pals[pct].b=calloc(size,sizeof(unsigned char))))panic("out of memory for palette");
  return(pct++);
}

void palette_set(int h,int index,unsigned long col)
{
  if(h>=pct)panic("bad palette in set");
  pals[h].b[index]+=col & 0xff;
  col>>=8;
  pals[h].g[index]+=col & 0xff;
  col>>=8;
  pals[h].r[index]+=col;
}

void palette_reset()
{
  int i;
  for(i=0;i<pct;i++)
	{
	  free(pals[i].r);
	  free(pals[i].g);
	  free(pals[i].b);
	}
  pct=0;
}

void palette_speed(int h,int s)
{
  if(h>=pct)panic("bad palette in speed");
  pals[h].speed=s;
}

void palette_gen(int h,unsigned long fore,unsigned long back,
				 long up,long down,long offset)
{
  unsigned long i,pos,size,fp,bp;
  unsigned short br,bg,bb,fr,fg,fb;
  unsigned short red,green,blue;

  if(h>=pct)panic("bad palette in palettegen");
  if(offset+up+down>100)panic("offset+up+down>100% in palettegen");

  br=PALRED(back);
  bg=PALGREEN(back);
  bb=PALBLUE(back);

  fr=PALRED(fore);
  fg=PALGREEN(fore);
  fb=PALBLUE(fore);

  /* scale up */
  up *= 100L;
  down *= 100L;
  offset *= 100L;

  size=pals[h].size;
  for(i=0;i<size;i++)
	{
	  /* work out how far through the palette we are, as
		 a %tage */

	  graph_check();

	  pos=i*10000L / size;

	  /* How much fg/bg is there? */

	  if(pos<offset)
		fp=0L;
	  else if(pos<offset+up)
		fp=(10000L*(pos-offset))/up;
	  else if(pos<offset+up+down)
		fp=10000L-(10000L*(pos-offset-up))/down;
	  else
		fp=0L;

	  bp=10000L-fp;

	  /* work out the colours */

	  red=	((long)fr*fp)/10000L + ((long)br*bp)/10000L;
	  green=((long)fg*fp)/10000L + ((long)bg*bp)/10000L;
	  blue=	((long)fb*fp)/10000L + ((long)bb*bp)/10000L;

	  pals[h].r[i]+=(char)red;
	  pals[h].g[i]+=(char)green;
	  pals[h].b[i]+=(char)blue;
	}
}


/*
  Assemble the various palettes into a single
  palette, and increment the pointers
  */

void palette_run()
{
  long h,i,p,offset,palsize;
  long r,g,b;
  unsigned char *rp,*gp,*bp;
  
  struct palette *thispal;
  
  palette_init();
  if(!pct)return;

  for(h=0L;h<(long)pct;h++)
	{
	  thispal=&(pals[h]);

	  offset=thispal->ct;
	  palsize=thispal->size;
	  rp=thispal->r;gp=thispal->g;bp=thispal->b;
	  
	  /* add in the colours for this palette */
	  for(i=0;i<size;i++)
		{
		  /* get the colour index we're going to use */
		  p=(i+offset)%palsize;

		  /* and add the colour in */
		  pal[i][0]+=rp[p];
		  pal[i][1]+=gp[p];
		  pal[i][2]+=bp[p];
		}

	  /* do the speed for this palette */
	  offset+=thispal->speed;
	  if(offset>=palsize)offset-=palsize;
	  if(offset<0)offset+=palsize;
	  thispal->ct=offset;
	}

  for(i=0;i<size;i++)
	{
	  r=pal[i][0];
 	  if(r>255)r=255;
 	  g=pal[i][1];
 	  if(g>255)g=255;
 	  b=pal[i][2];
 	  if(b>255)b=255;

	  if(fade)
		{
		  r=(100-fade)*r+fadered*fade;
		  r/=100;
		  g=(100-fade)*g+fadegreen*fade;
		  g/=100;
		  b=(100-fade)*b+fadeblue*fade;
		  b/=100;
		}
	  pal[i][0]=r;pal[i][1]=g;pal[i][2]=b;
	}
  if(fade && fade<100)
	{
	  fade+=fadespeed;
	  if(fade>=100)
		{
		  fade=100;fadespeed=0;
		}
	  else if(fade<=0)
		{
		  fade=0;fadespeed=0;
		}
	}
}

void palette_startfadeout(int r,int g,int b,int s)
{
  static int firsttime=1;

  if(firsttime)
	firsttime=0;
  else
	{
	  fadespeed=s;
	  fadered=r;
	  fadegreen=g;
	  fadeblue=b;
	  fade=1;
	  while(fadespeed)graph_check();
	}
}

void palette_startfadein(int s)
{
  fadespeed= -s;
  fade=99;
  while(fadespeed)graph_check();
}

int palette_isfadedone()
{
  return(!fadespeed);
}


void L_NewPalette()
{
  Push(palette_new(Pop()));
}

void L_PaletteSet()
{
  long h,i;
  unsigned long col;

  col=(unsigned long)Pop();
  i=Pop();
  h=Pop();

  palette_set(h,i,col);
}

void L_PaletteReset()
{
  palette_reset();
}

void L_GenPalette()
{
  long h,up,down,offset;
  unsigned long fg,bg;

  offset=Pop();down=Pop();up=Pop();
  bg=(unsigned long)Pop();
  fg=(unsigned long)Pop();
  h=Pop();

  palette_gen(h,fg,bg,up,down,offset);
}

  
void L_PaletteSpeed()
{
  long h,s;
  s=Pop();h=Pop();
  palette_speed(h,s);
}

void L_MkRGB()
{
  long r,g,b;
  b=Pop();g=Pop();r=Pop();
  Push(PALRGB(r,g,b));
}

void L_MkHSV()
{
  long r,g,b,s,v,p,q,t,i,f,h;
  
  v=(long)Pop();s=(long)Pop();h=(long)Pop();

  if(!s)
	{
	  Push(PALRGB(v,v,v));
	  return;
	}

  h %= 360L;
  h*=256L;	/* scale up h, so we work in fixed point */

  h/=60L;

  i=h/256L;
  i*=256L;
  
  f=h-i;

  p=v*(256L-s)/256L;
  q=v*(256L-(s*f)/256L)/256L;
  t=v*(256L-(s*(256L-f))/256L)/256L;

  switch(i/256L)
	{
	case 0L:r=v;g=t;b=p;break;
	case 1L:r=q;g=v;b=p;break;
	case 2L:r=p;g=v;b=t;break;
	case 3L:r=p;g=q;b=v;break;
	case 4L:r=t;g=p;b=v;break;
	case 5L:r=v;g=p;b=q;break;
	}
  Push(PALRGB(r,g,b));
}

