/*
	Yacc file for the springlang compiler
*/

%{

static char rcsid[]="$Id: parser.y,v 1.4 1994/09/28 21:44:00 white Exp white $";

#define	YYDEBUG	1
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include "compiler.h"
#include "opcodes.h"
#include "spring.h"

extern int lines,show_out;
LIST *args=NULL,*locals=NULL,*functions=NULL;
LIST *functab[MAXFUNCS];
LIST *imagefunctab[MAXFUNCS];
static int argct=0;
static int localct=0;
static int got_a_return=0;

int funcct=0,ifuncct=0;

LIST *addlist(LIST *,char *,int n,...);
void deletelist(LIST *);
LIST *find(LIST *,char *);
NODE *join(NODE *,NODE *);
NODE *mknod(int);
NODE *joinexpr(NODE *,NODE *);
NODE *do_switch(NODE *,NODE *,NODE *);
LIST *mkvar(char *id);
void do_func(char *name,int numargs,NODE *block,int doesreturn);
NODE *do_block(NODE *block);
NODE *do_null(void);
NODE *do_if(NODE *,NODE *,NODE *);
NODE *do_for(NODE *,NODE *,NODE *,NODE *);
NODE *do_while(NODE *,NODE *);
NODE *do_assign(char *id,NODE *expr,int dodrop);
NODE *do_return(NODE *);
NODE *do_label(char *);
NODE *do_goto(char *);
NODE *do_case(long,NODE *);
NODE *do_funccall(char *name,NODE *exprlist,int inexpr);
NODE *do_bin(NODE *a,int type,NODE *b);
NODE *do_un(NODE *a,int type);
NODE *do_num(long n);
NODE *do_getvar(char *n);

%}

%start file

%union
{
	char str[64];
	long num;
	NODE *node;
}

%token	OR AND NOTEQUAL GTEQUAL LTEQUAL UMINUS FUNCTION
%token	SWITCH DEFAULT CASE FOR WHILE IF ELSE GOTOTOK
%token	RETURN NOT EQUAL VOID LOCAL
%token	ENDOFFILE IMAGE
 
%token	<str>	ID
%token	<num>	NUMBER
%token	<num>	STRING

%type	<node>	block
%type	<node>	stmtlist
%type	<node>	stmt
%type	<node>	ifstmt
%type	<node>	whilestmt
%type	<node>	returnstmt
%type	<node>	gotostmt
%type	<node>	teststmt
%type	<node>	forstmt
%type	<node>	assignstmt
%type	<node>	labelstmt
%type	<node>	funcstmt
%type	<node>	switchstmt
%type	<node>	caselist
%type	<node>	caseitem
%type	<node>	exprlist
%type	<node>	expr

%right '='
%left OR
%left AND
%nonassoc EQUAL NOTEQUAL '<' '>' GTEQUAL LTEQUAL
%left '%'
%left '+' '-'
%left '*' '/'
%left UMINUS

%%

file		:		funclist ENDOFFILE
					{
						YYACCEPT;
					}
			;

funclist	:		func
			|		funclist func
			;

func		:		VOID ID '(' arglist ')' block
					{
						/* arglist is is the list "args" */
						do_func($2,argct,$6,0);
						deletelist(args);argct=0;args=NULL;
						deletelist(locals);localct=0;locals=NULL;
					}	
			|		VOID ID '(' arglist ')' ';'
					{
						do_func($2,argct,NULL,0);
						deletelist(args);argct=0;args=NULL;
					}
			|		IMAGE ID '(' ')' block
					{
					  	/* generate an image function */
					  	do_func($2,0,$5,2);
						deletelist(args);argct=0;args=NULL;
						deletelist(locals);localct=0;locals=NULL;
					}

			|		IMAGE ID '(' ')' ';'
					{
					  /* declare an image function */
					  do_func($2,0,NULL,2);
					}
			|		ID '(' arglist ')' block
					{
						do_func($1,argct,$5,1);
						deletelist(args);argct=0;args=NULL;
						deletelist(locals);localct=0;locals=NULL;
					}
			|		ID '(' arglist ')' ';'
					{
						do_func($1,argct,NULL,1);
						deletelist(args);argct=0;args=NULL;
					}
			;

arglist		:		arg
			|		arglist ',' arg
			|
			;

arg			:		ID
					{
						/* add an argument to the global argument list */
						if(find(args,$1))panic("%s is already an arg",$1);
						args=addlist(args,$1,1,argct++);
					}
			;
			
block		:		'{' declist stmtlist '}'
					{
						/* generate a "block" statement node */
						$$=do_block($3);
					}
			|		'{' '}'
					{
						/* generate a "null" statement node */
						$$=do_null();
					}
			;

declist		:		decl
			|		declist decl
			|
			;

decl		:		LOCAL localidlist ';'
		;

localidlist	:	localid
			|	localidlist ',' localid
			;

localid		:	ID
				{
					mkvar($1);
				}
			;

stmtlist	:	stmt
			|	stmtlist stmt
				{
					$$=join($1,$2);
				}
			;

stmt		:	ifstmt
			|	whilestmt
			|	teststmt
			|	assignstmt
			|	funcstmt
			|	returnstmt
			|	gotostmt
			|	forstmt
			|	labelstmt
			|	switchstmt
			|	block
			;

assignstmt	:	ID '=' expr ';'
				{
					$$=do_assign($1,$3,1);
				}
			;

funcstmt	:	ID '(' exprlist ')' ';'
				{
					$$=do_funccall($1,$3,0);
				}
			;

switchstmt	:	SWITCH '(' expr ')' '{'  caselist DEFAULT ':'  stmt '}'
			{
				$$=do_switch($3,$6,$9);
			}
			|	SWITCH '(' expr ')' '{'  caselist '}'
			{
				panic("no default item in switch");
			}
			;

caselist	:	caseitem
			|	caselist caseitem
				{ $$=join($1,$2);}
			;

caseitem	:	CASE NUMBER ':' stmt
				{
					$$=do_case($2,$4);
				}
			;

forstmt		:	FOR '(' stmt expr ';' expr ')' stmt
				{
				  $6->value.str[0] |= FUNCB_ISSTMT;	/* turn that expr into a statement */
				  $$=do_for($3,$4,$6,$8);
				}
			;

labelstmt	:	ID ':'
				{
					$$=do_label($1);
				}
			;

gotostmt	:	GOTOTOK ID ';'
				{
					$$=do_goto($2);
				}
			;

ifstmt		:	IF '(' expr ')' stmt ELSE stmt
			{
				$$=do_if($3,$5,$7);
			}
			
teststmt	:	IF '(' expr ')' stmt
			{
				$$=do_if($3,$5,NULL);
			}
			;

whilestmt	:	WHILE '(' expr ')' stmt
			{
				$$=do_while($3,$5);
			}
			;

returnstmt	:	RETURN expr ';'
			{
				$$=do_return($2);
			}
			|	RETURN ';'
			{
				$$=do_return(NULL);
			}      
			;

exprlist	:
				{$$=NULL;}
			|	expr
			|	exprlist ',' expr
				{ $$=joinexpr($1,$3); }

			;
			
expr		:	ID '=' expr %prec '='
				{
					$$=do_assign($1,$3,0);
				}
			|	expr '+' expr
				{
					$$=do_bin($1,NT_ADD,$3);
				}
			|	expr '-' expr
				{
					$$=do_bin($1,NT_SUB,$3);
				}
			|	expr '*' expr
				{
					$$=do_bin($1,NT_MUL,$3);
				}
			|	expr '/' expr
				{
					$$=do_bin($1,NT_DIV,$3);
				}
			|	expr '>' expr
				{
					$$=do_bin($1,NT_GREATER,$3);
				}
			|	expr '<' expr
				{
					$$=do_bin($1,NT_LESS,$3);
				}
			|	expr EQUAL expr
				{
					$$=do_bin($1,NT_EQUAL,$3);
				}
			|	expr NOTEQUAL expr
				{
					$$=do_bin($1,NT_NOTEQUAL,$3);
				}
			|	expr GTEQUAL expr
				{
					$$=do_bin($1,NT_GTEQUAL,$3);
				}
			|	expr LTEQUAL expr
				{
					$$=do_bin($1,NT_LTEQUAL,$3);
				}
			|	expr OR expr
				{
					$$=do_bin($1,NT_OR,$3);
				}
			|	expr AND expr
				{
					$$=do_bin($1,NT_AND,$3);
				}
			|	expr '%' expr
				{
					$$=do_bin($1,NT_MOD,$3);
				}
			|	'-' expr %prec UMINUS
				{
					$$=do_un($2,NT_NEGATE);
				}
			|	NOT expr %prec UMINUS
				{
					$$=do_un($2,NT_NOT);
				}
			|	'(' expr ')'
				{
					$$=$2;
				}
			|	NUMBER
				{
					$$=do_num($1);
				}
			|	ID
				{
					$$=do_getvar($1);
				}
			|	ID '(' exprlist ')'
				{
					$$=do_funccall($1,$3,1);
				}
			;
%%



/****************************************************************************
	Data structure handling utilities
 ****************************************************************************/

LIST *addlist(LIST *list,char *str,int n,...)
{
	int i;

	va_list ap;
	LIST *new=malloc(sizeof(LIST));
	va_start(ap,n);
	if(n>LISTNUMS)panic("too many numbers");
	if(!new)panic("no memory to add to list!");
	strcpy(new->str,str);

	for(i=0;i<n;i++)new->nums[i]=va_arg(ap,int);
	
	new->next=list;
	new->flags=0L;
	return(new);
}

LIST *find(LIST *list,char *str)
{
	for(;list;list=list->next)if(!strcmp(str,list->str))return(list);
	return(NULL);
}

void deletelist(LIST *list)
{
	LIST *p,*q;
	if(!list)return;
	for(p=list;p;)
	{
		q=p->next;
		free(p);
		if(!(p=q))break;
	}
}

NODE *mknod(int type)
{

	int i;
	NODE *new=malloc(sizeof(NODE));
	if(!new)panic("No memory to make a new node!");

	for(i=0;i<5;i++)new->nodes[i]=NULL;
	new->value.str[0]='\0';
	new->value.num=0;
	new->value.num2=0;
	new->value.num3=0;
	new->value.num4=0;
	new->type=type;
	new->numnodes=0;
	new->line=lines;
	return(new);
}

void deletetree(NODE *t)
{
	int i;
	if(!t)return;

	for(i=0;i<5;i++)
	{
		if(t->nodes[i]==t)panic("Tree recursion! Type %d\n",t->type);
		if(t->nodes[i])
		{
			deletetree(t->nodes[i]);
			t->nodes[i]=NULL;
		}
	}
	free(t);
}

NODE *join(NODE *a,NODE *b)
{
	NODE *t;

	if(!a)return(b);
	if(!b)return(a);
	for(t=a;t->nodes[1];t=t->nodes[1]);

	t->nodes[1]=b;return(a);
}

NODE *joinexpr(NODE *a,NODE *b)
{
	NODE *t;

	if(!a)return(b);
	if(!b)return(a);
	for(t=a;t->nodes[4];t=t->nodes[4]);

	t->nodes[4]=b;return(a);
}

/****************************************************************************
	Structure generation code
 ****************************************************************************/

void clearall()
{
	lines=0;funcct=0;localct=0;
	deletelist(args);args=NULL;
	deletelist(locals);locals=NULL;
}

/* declare a function, with or without code. */

void do_func(char *name,int numargs,NODE *block,int doesreturn)
{
  LIST *f;
  int isimage=0;

  if(doesreturn==2)
	{
	  doesreturn=0;isimage=1;
	  if(numargs)panic("Image function %s can't take arguments, sorry.",name);
	}


  if(f=find(functions,name))
	{
	  /* Righto, this function is a either a declaration or
		 some stupid kennet has defined it twice. */

	  if(f->nums[2]>=0)	/* is it already defined? */
		panic("errr.. I've already done function %s once, thanks.");
	  else if(!block)
		panic("look, are you going to get round to actually *defining* %s?",name);
	}

  /* If we need to set up a new function entry, do so. */

  if(!f)
	{
	  f=functions=addlist(functions,name,3,funcct,doesreturn,-1);
	  if(funcct==MAXFUNCS)
		panic("too many functions - the last straw was %s",name);
	  functab[funcct++]=f;
	  /* if an image function, register it */
	  if(isimage)
		{
		  imagefunctab[ifuncct++]=f;
		}
	}

  /* Groovy. We now have a function slot. All we need to do is
	 deal with any function body. */

  if(block)
	{
	  /* There's a body, so deal with it */
  
	  /* Did we get a return in the block? If so, make sure we are supposed
		 to return values. If not, make sure we're *not* supposed to return
		 values */

	  if(got_a_return)
		{
		  got_a_return=0;
		  if(!doesreturn)panic("cannot return a value from image/void function %s",name);
		}
	  else
		{
		  if(doesreturn)panic("must return a value from function %s",name);
		}

	  dprintf("Compiled %s to %ld\n",name,getcompilerpos());
	  f->nums[2]=getcompilerpos();	/* set up the offset */


	  if(numargs)
		{
		  /* OK, there are some arguments to deal with.
			 Generate the instructions to unstack them into the arguments
			 array */
		  output(numargs);output(OP(OP_GETARGS));
		}

	  /* Then output and delete the code */

	  if(show_out)
		printf("VM code for function %s. (%d args, returns %s)\n",
			   name,numargs,doesreturn?"a value":
							isimage?"image":"nothing");

	  outputtree(block);
	  deletetree(block);
	  outputreturn();
	  flushbuffer();
	  localct=0;
	}
}

NODE *do_block(NODE *block)
{
	NODE *n,*s;
	n=mknod(NT_BLOCK);
	s=mknod(NT_STATEMENT);
	n->nodes[0]=block;
	s->nodes[0]=n;
	return(s);
}

NODE *do_null(void)
{
	NODE *n,*s;
	n=mknod(NT_NULL);
	s=mknod(NT_STATEMENT);
	s->nodes[0]=n;
	return(s);
}

NODE *do_switch(NODE *expr,NODE *caselist,NODE *def)
{
	NODE *s,*n;

	/* make the main switch node */

	s=mknod(NT_STATEMENT);
	n=mknod(NT_SWITCH);
	n->nodes[0]=expr;
	n->nodes[1]=caselist;
	n->nodes[2]=def;
	s->nodes[0]=n;
	return(s);
}

NODE *do_case(long number,NODE *stmt)
{
	NODE *n;

	n=mknod(NT_CASE);
	n->value.num=number;
	n->nodes[0]=stmt;
	return(n);
}

NODE *do_while(NODE *expr,NODE *stmt)
{
	NODE *n,*s;

	n=mknod(NT_WHILE);
	s=mknod(NT_STATEMENT);

	n->nodes[0]=expr;
	n->nodes[1]=stmt;
	s->nodes[0]=n;
	return(s);
}


NODE *do_for(NODE *start,NODE *cond,NODE *end,NODE *body)
{
	NODE *s,*n;
	s=mknod(NT_STATEMENT);
	n=mknod(NT_FOR);
	n->nodes[0]=start;
	n->nodes[1]=cond;
	n->nodes[2]=end;
	n->nodes[3]=body;
	s->nodes[0]=n;
	return(s);
}

NODE *do_if(NODE *expr,NODE *t,NODE *f)
{
	NODE *n,*s;
	n=mknod(NT_IF);
	s=mknod(NT_STATEMENT);

	n->nodes[0]=expr;
	n->nodes[1]=t;
	n->nodes[2]=f;
	s->nodes[0]=n;
	return(s);
}

NODE *do_return(NODE *expr)
{
	NODE *n,*s;

	/* NOTE: a return instruction should not be generated from
	inside a expression block. */

	n=mknod(NT_RETURN);
	s=mknod(NT_STATEMENT);
	n->nodes[0]=expr;	/* may well be null! */
	s->nodes[0]=n;

	/* Here we tell the compiler that a return appeared somewhere in the
	   block - but only if the return actually returns a value */

	if(expr)got_a_return=1;

	return(s);
}

NODE *do_label(char *nam)
{
	NODE *n,*s;

	n=mknod(NT_LABEL);
	s=mknod(NT_STATEMENT);
	strcpy(n->value.str,nam);
	s->nodes[0]=n;
	return(s);
}

NODE *do_goto(char *nam)
{
	NODE *n,*s;

	n=mknod(NT_GOTO);
	s=mknod(NT_STATEMENT);
	strcpy(n->value.str,nam);
	s->nodes[0]=n;
	return(s);
}

LIST *mkvar(char *id)
{
	if(find(locals,id))
		panic("variable already declared - %s",id);

	if(localct==MAXVARS)panic("too many local variables");
	return(locals=addlist(locals,id,1,localct++));
}

NODE *do_assign(char *id,NODE *expr,int dodrop)
{
	NODE *n,*s;
	LIST *var;
	int nodetype=NT_ASSIGN;

	if(var=find(args,id))nodetype=NT_ASSIGNARG;
	else if(!(var=find(locals,id)))panic("variable %s not found",id);

	n=mknod(nodetype);
	s=mknod(NT_STATEMENT);

	n->nodes[0]=expr;
	n->value.num=var->nums[0];
	s->nodes[0]=n;
	n->value.str[0] = dodrop ? 1 : 0;
	return(s);
}


NODE *do_funccall(char *nam,NODE *exprlist,int inexpr)
{
	LIST *f;
	int doesreturn=0;
	int isvarargs=0;
	NODE *n,*s;

	if(f=find(functions,nam))
	  {
		n=mknod(NT_FUNCCALL);

		n->nodes[0]=exprlist;
		n->value.num=f->nums[0];
		if(!(doesreturn=f->nums[1]) && inexpr)
		  panic("sorry, can't use void function %s in an expression, pal.",nam);
	  }
	else
	  {
		extern LIST *getlibcall(char *,NODE *);

		if(!(f=getlibcall(nam,exprlist)))
		  panic("function not found, either in your code or the library - %s",nam);

		n=mknod(NT_LIBCALL);
		n->nodes[0]=exprlist;
		n->value.num=f->nums[0];
		doesreturn=f->nums[2];
		isvarargs = f->flags & LF_VARARGS;
	  }

	s=mknod(NT_STATEMENT);
	s->nodes[0]=n;

	/* The first byte in the string field has certain special flag values,
	   telling the code generator about the function's peculiarities.
	   This is where we set them.		 */

	n->value.str[0]=0;

    if(f)
	  {
		n->value.str[0] |= isvarargs ? FUNCB_VARARGS : 0;
		n->value.str[0] |= doesreturn ? FUNCB_RETURNS : 0;
	  }

	if(!inexpr)
	  {
	  /* we'll need a OP_DROP if this AND FUNCB_RETURNS are true */
	  n->value.str[0] |= FUNCB_ISSTMT;
	  }
	return(s);
}

NODE *do_bin(NODE *a,int type,NODE *b)
{
	NODE *n;
	n=mknod(type);

	n->nodes[0]=a;
	n->nodes[1]=b;
	return(n);
}

NODE *do_un(NODE *a,int type)
{
	NODE *n;
	n=mknod(type);
	n->nodes[0]=a;
	return(n);
}

	

NODE *do_num(long nn)
{
	NODE *n;
	n=mknod(NT_NUMBER);
	n->value.num=nn;
	return(n);
}

NODE *do_getvar(char *name)
{
	LIST *var;
	NODE *n;
	int nodetype;
	

	if(var=find(args,name))
		nodetype=NT_GETARG;
	else if(var=find(locals,name))
		nodetype=NT_GETVAR;
	else panic("variable %s not found",name);

	n=mknod(nodetype);
	n->value.num=var->nums[0];
	return(n);
}
	

