#include <stdio.h>
#include <dos.h> 

#include "spring.h"
#include "graphics.h"
#include "palette.h"   
#include "buffer.h"


/*
	The machine-dependent graphics (PC vga, 320x200)
*/

void interrupt (*oldvect)()=NULL;

int palette_size=0;
static int dis_x=0,dis_y=0;
int win_x=0,win_y=0;
int win_scale=1;

extern char **argvec;

int go_tick=0;
int timing=1;
unsigned long ticks=0L;
void interrupt graph_tsig()
{
	disable();
	ticks++;
	go_tick=1;
	oldvect();
	enable();
}

void graph_init()
{
	union REGS regs;

	regs.h.ah=0x00;	/* set video mode */
	regs.h.al=0x13;	/* 320 x 200, 256 colour */
	int86(0x10,&regs,&regs);

	dis_x=320;win_x=dis_x/win_scale;
	dis_y=200;win_y=dis_y/win_scale;
		palette_size=256;
	
	if((dis_x % win_x) || (dis_y % win_y))
	{
		printf("Display size: %d,%d\n",dis_x,dis_y);
		panic("Display size must be exact multiple of window size");
	}

        oldvect=getvect(0x08);
	setvect(0x08,graph_tsig);
	atexit(graph_shutdown);
}

#define	COLMASK	0x3f

void graph_palette(short *p)
{
  union REGS regs;
  struct SREGS sregs;

  char cols[256][3];

  int i;
  for(i=0;i<palette_size;i++)
  {
	cols[i][0]=(*p++ >> 2)&0x3f;
	cols[i][1]=(*p++ >> 2)&0x3f;
	cols[i][2]=(*p++ >> 2)&0x3f;
  }

  regs.h.ah=0x10;
  regs.h.al=0x12;
  regs.x.bx=0;
  regs.x.cx=palette_size;

  regs.x.dx=FP_OFF(cols);
  sregs.es=FP_SEG(cols);

  int86x(0x10,&regs,&regs,&sregs);
}

void graph_show()
{
	extern char *bufferdata;
	static char *line=NULL;
	char *p,*q;
	int x,y,i,yd;
	/* VGA 8-bit display address */
	char far *screen=MK_FP(0xa000,0x0000);

	if(!line)
	{
		if(!(line=malloc(dis_x)))
			panic("unable to allocate for scanline");
	}

	for(yd=y=0;y<win_y;y++)
	{
		p=bufferdata+win_x*y;
		for(q=line,x=0;x<win_x;x++)
		{
			for(i=0;i<win_scale;i++)*q++ = *p;
			p++;
		}
		for(i=0;i<win_scale;i++)
			memcpy(screen+dis_x*yd++,line,dis_x);
	}
	buffer_clear();
}

void graph_shutdown()
{
	union REGS regs;

	if(oldvect)
	{
		regs.h.ah=0x00;
		regs.h.al=0x02;
		int86(0x10,&regs,&regs);
		setvect(0x08,oldvect);
		oldvect=NULL;
	}
}

int holding=0;	/* TRUE if the interpreter is halted (by the H key) */

char *quitmsgs[]=
{
	"You quit, you fiend",
	"You chickened out!",
	"You decided to return to reality",
	"You got bored",
	"Program aborted by the forces of Greyface",
	"Program aborted by CIA mind-control of user",
	"Program aborted for no very good reason",
	"You decided something else was prettier",
	"You felt seasick, so you quit",
	"You couldn't take any more",
};
#define	QUITMSGS	(sizeof(quitmsgs)/sizeof(char *))

void graph_check(void)
{
	int key;
	if(go_tick)
	{
		go_tick=0;
		if(key=bioskey(1))
		{
			switch(key)
			{
				case 8051:
					holding=1;break;
				case 8807:
					holding=1;break;
				default:
					graph_shutdown();
					panic(quitmsgs[rand()%QUITMSGS]);
			}
		}
		palette_run();
		graph_palette(palette_get());
	}
}


void graph_setsize(short s)
{
	if(s>0)win_scale=s;
	if(s>2)panic("only -r2 is valid in this version.");
}

void graph_settiming(short t)
{
	if(t!=1)panic("timing not yet implemented");
	timing=t;
}
