/*
	Wave stuff
*/

#include "spring.h"
#include "math.h"
#include "buffer.h"
#include "graphics.h"
#include "waves.h"
#include "inter.h"

static char rcsid[]="$Id: waves.c,v 1.2 1994/09/17 21:51:09 white Exp $";

struct wave
{
  long size;
  short *vals;
};

static struct wave waves[MAXWAVES];
static long wavct=0;

long wave_new(long size)
{
  if(wavct==MAXWAVES)panic("too many waves");

  if(!(waves[wavct].vals=(short *)
	   calloc(waves[wavct].size=size,
			  sizeof(short))))
	panic("not enough memory for wave");

  return(wavct++);
}

long wave_new_table(long size,short *tab)
{
  if(wavct==MAXWAVES)panic("too many waves");

  waves[wavct].size=size;
  waves[wavct].vals=tab;

  return(wavct++);
}

void wave_set(long wavindex,long index,short val)
{
  if(wavindex>=wavct)panic("bad wave in set %d",wavindex);
  waves[wavindex].vals[index]=(short)val;
}

void wave_drawh(long wavindex,long x,long y,long len,long offset,
			   long scale,long intensity,long paloffset)
{
  unsigned long size,i,palsize,start,index;
  unsigned long col;
  short *data;

  palsize=graph_psize();

  if(wavindex>=wavct)panic("bad wave in draw %d",wavindex);

  if(y>graph_ysize())return;
  if(x+len > graph_xsize())len=graph_xsize()-x;

  data=waves[wavindex].vals;
  if(!(size=waves[wavindex].size))panic("funny wave %d",wavindex);

  /* work out the start */

  start=(offset*size)/100L;

  for(i=0;i<len;i++)
	{
	  graph_check();
	  /* get index into table */
	  index=i*size*scale;
	  index=index/(len*100L)+start;
	  index %= size;


	  col=data[index]*intensity; /* data in the wave is 0-32767 */
	  /* convert to a colour */
	  col/=100L*(32768L/palsize);
	  /* add the palette offset and take the mod */
	  col=(col+paloffset)%palsize;

	  /* plot! */

	  buffer_point(x++,y,col);
	}
}
	  
void wave_drawv(long wavindex,long x,long y,long len,long offset,
			   long scale,long intensity,long paloffset)
{
  unsigned long size,i,palsize,start,index;
  unsigned long col;
  short *data;

  palsize=graph_psize();

  if(wavindex>=wavct)panic("bad wave in draw %d",wavindex);

  if(x>graph_xsize())return;
  if(y+len > graph_ysize())len=graph_ysize()-y;

  data=waves[wavindex].vals;
  if(!(size=waves[wavindex].size))panic("funny wave %d",wavindex);

  /* work out the start */

  start=(offset*size)/100L;

  for(i=0;i<len;i++)
	{
	  graph_check();
	  /* get index into table */
	  index=i*size*scale;
	  index=index/(len*100L)+start;
	  index %= size;

	  col=data[index]*intensity; /* data in the wave is 0-32767 */
	  /* convert to a colour */
	  col/=100L*(32768L/palsize);
	  /* add the palette offset and take the mod */
	  col=(col+paloffset)%palsize;
	  
	  /* plot! */

	  buffer_point(x,y++,col);
	}
}

/*
  Set up the internal waves
  */

static long SawWave,SineWave;
void wave_init()
{
  long i;

  extern short wSinTab[];	/* from math.h */
  SineWave=wave_new_table(4096,wSinTab);
  SawWave=wave_new(4096);

  for(i=0;i<2048;i++)
	wave_set(SawWave,i,i*(32767/2048));
  for(i=2048;i<4096;i++)
	wave_set(SawWave,i,(4096-i)*(32767/2048));
}

void wave_radial(long handle,long x,long y,
long size,long offset,long scale,long intensity)
{
	short *data;
	long palsize,nom,denom,wavsize,cdenom;
	long xp,yp,xs,ys;
	register long cx,cy,dist,col;

	palsize=graph_psize();

	if(handle>=wavct)panic("bad wave in radial %d",handle);
	wavsize=waves[handle].size;
	data=waves[handle].vals;
	
	/* work out ratios */

	nom=wavsize*scale;
	denom=size*100L;

	cdenom=100L*(32768L/palsize);
	
	xs=x+size;ys=y+size;
	xp=x-size;yp=y-size;
	
	if(xs>graph_xsize())xs=graph_xsize();
	if(ys>graph_ysize())ys=graph_ysize();
	if(xp<0)xp=0;
	if(yp<0)yp=0;

	offset *=wavsize;
	offset /=100;
	
	/* and looooooop */

	for(cx=xp;cx<xs;cx++)
	{
		for(cy=yp;cy<ys;cy++)
		{
			graph_check();
			dist=math_dist(cx,cy,x,y);

			dist*=nom;
			dist/=denom;

			col=(long)(data[(dist+offset)%wavsize]) * intensity;
			buffer_point(cx,cy,col/cdenom);
		}
	}
}
	
/*  User library calls 
 */

void L_HSineLine()
{
  long paloffset,intensity,scale,offset,len,x,y;

  paloffset=Pop();
  intensity=Pop();
  scale=Pop();
  offset=Pop();
  len=Pop();
  y=Pop();
  x=Pop();

  wave_drawh(SineWave,x,y,len,offset,scale,intensity,paloffset);
}

void L_VSineLine()
{
  long paloffset,intensity,scale,offset,len,x,y;

  paloffset=Pop();
  intensity=Pop();
  scale=Pop();
  offset=Pop();
  len=Pop();
  y=Pop();
  x=Pop();

  wave_drawv(SineWave,x,y,len,offset,scale,intensity,paloffset);
}

void L_HSawLine()
{
  long paloffset,intensity,scale,offset,len,x,y;

  paloffset=Pop();
  intensity=Pop();
  scale=Pop();
  offset=Pop();
  len=Pop();
  y=Pop();
  x=Pop();

  wave_drawh(SawWave,x,y,len,offset,scale,intensity,paloffset);
}

void L_VSawLine()
{
  long paloffset,intensity,scale,offset,len,x,y;

  paloffset=Pop();
  intensity=Pop();
  scale=Pop();
  offset=Pop();
  len=Pop();
  y=Pop();
  x=Pop();

  wave_drawv(SawWave,x,y,len,offset,scale,intensity,paloffset);
}

void L_HWaveLine()
{
  long paloffset,intensity,scale,offset,len,x,y,w;

  paloffset=Pop();
  intensity=Pop();
  scale=Pop();
  offset=Pop();
  len=Pop();
  y=Pop();
  x=Pop();
  w=Pop();

  wave_drawh(w,x,y,len,offset,scale,intensity,paloffset);
}

void L_VWaveLine()
{
  long paloffset,intensity,scale,offset,len,x,y,w;

  paloffset=Pop();
  intensity=Pop();
  scale=Pop();
  offset=Pop();
  len=Pop();
  y=Pop();
  x=Pop();
  w=Pop();

  wave_drawv(w,x,y,len,offset,scale,intensity,paloffset);
}

void L_WaveNew()
{
  Push(wave_new(Pop()));
}

void L_WaveSet()
{
  long val,i,w;

  val=Pop();i=Pop();w=Pop();
  wave_set(w,i,val);
}

void L_Radial()
{
	long intensity,scale,offset,size,y,x,handle;

	intensity=Pop();scale=Pop();offset=Pop();size=Pop();
	y=Pop();x=Pop();handle=Pop();

	wave_radial(handle,x,y,size,offset,scale,intensity);
}

