#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#include "spring.h"
#include "graphics.h"
#include "palette.h"

static char rcsid[]="$Id: xgraphics.c,v 1.2 1994/09/17 21:51:09 white Exp $";

/*
	The machine-dependent graphics (X windows)
*/

#define	XSIZE	-1
#define	YSIZE	-1

/* time between ticks in usec */
#define	TIMEOUT	50000L

/* number of ticks between event checks */
#define	ECHECKTICKS	10000

int holding=0,timeout=TIMEOUT;
int palette_size=0;
int win_x=XSIZE,win_y=YSIZE;
extern char **argvec;
static int dis_x=0,dis_y=0;

Display *display=NULL;
int screen;
Window window;
Visual *visual;
Colormap cmap,orig_cmap;
XColor color;
GC gc;

static int do_tick=0;

void maketimer()
{
  static struct itimerval tim=
  {
	{0L,TIMEOUT},
	{0L,TIMEOUT}
  };

  struct itimerval otim;

  tim.it_interval.tv_usec=timeout;
  tim.it_value.tv_usec=timeout;

  setitimer(ITIMER_REAL,&tim,&otim);
}

void graph_tsig()
{
	signal(SIGALRM,graph_tsig);
	do_tick=1;
}

void graph_check()
{
	/* only check for events every now and then,
	because it's a real hog */

	static int echeck=ECHECKTICKS;
	XEvent e;

	if(do_tick)
	{
		palette_run();
		graph_palette(palette_get());
		if(--echeck)
		{
			echeck=ECHECKTICKS;
			if(XCheckMaskEvent(display,ButtonPressMask,&e))
			{
				panic("quit");
			}
		}
		do_tick=0;
	}
}

void graph_setsize(short x,short y)
{
  win_x = (x>0)?x:XSIZE;
  win_y = (y>0)?y:YSIZE;
}

void graph_settiming(short timing)
{
  timeout=TIMEOUT*timing;
}

void graph_init()
{
  XSizeHints sizehints;
  XVisualInfo vTemplate,*visualList;
  int visualsMatched;
  XSetWindowAttributes atts;
  XEvent report;
  display=XOpenDisplay("");
  if(!display)panic("oops - no display!");
  screen=DefaultScreen(display);
  window=XDefaultRootWindow(display);

  dis_x=DisplayWidth(display,screen);
  dis_y=DisplayHeight(display,screen);

  /*
		If the user doesn't specify a window size,
		divide the display size down to a reasonable
		size
  */

  if(win_x<0)
  {
	win_x=dis_x;
	win_y=dis_y;
	
	while(win_x>512)
	{
	  win_x /= 2;
	  win_y /= 2;
	}
  }
		

  vTemplate.screen = screen;
  vTemplate.depth = 8;
  vTemplate.class = PseudoColor;
  visualList = XGetVisualInfo(display,VisualScreenMask|VisualClassMask|
							  VisualDepthMask,&vTemplate,&visualsMatched);

  if(!visualsMatched)panic("no suitable visual found");
  printf("found %d visuals\n",visualsMatched);
  visual=visualList[0].visual;

  sizehints.x=0;
  sizehints.y=0;
  sizehints.width=dis_x;
  sizehints.height=dis_y;
  sizehints.flags=USSize | USPosition;

  if((dis_x % win_x) || (dis_y % win_y))
	{
	  printf("Display size: %d,%d\n",dis_x,dis_y);
	  panic("Display size must be exact multiple of window size");
	}

  cmap=XCreateColormap(display,window,visual,AllocAll);
  orig_cmap=DefaultColormap(display,DefaultScreen(display));
  atts.colormap=cmap;
  atts.override_redirect=1;

  /* OK, it says in the book not to do this, but I can't figure
  out any alternative. */
  
  XInstallColormap(display,cmap);
  
  window=XCreateWindow(display,DefaultRootWindow(display),
					   sizehints.x,sizehints.y,sizehints.width,
					   sizehints.height,0,8,InputOutput,visual,
					   CWColormap|CWOverrideRedirect,&atts);

  XSetStandardProperties(display,window,argvec[0],argvec[0],None,argvec,1,&sizehints);


  color.flags = DoRed|DoBlue|DoGreen;
  palette_size=visual->map_entries;
  printf("Allocating %d colors\n",visual->map_entries);
  for(color.pixel=0;color.pixel<visual->map_entries;color.pixel++)
	{
	  color.red=0;color.green=0;color.blue=0;
	  XStoreColor(display,cmap,&color);
	}
  XFree((char *)visualList);
  XSelectInput(display,window,ExposureMask|ButtonPressMask);
  XMapWindow(display,window);
  gc=XCreateGC(display,DefaultRootWindow(display),0,NULL);

  /* Wait for the first Expose */

  XNextEvent(display,&report);

  signal(SIGALRM,graph_tsig);
  maketimer();
}

void graph_palette(short *p)
{
  int i;
  XColor colors[256];

  for(i=0;i<palette_size;i++)
	{
	  colors[i].red=256 * *p++;
	  colors[i].green=256 * *p++;
	  colors[i].blue=256 * *p++;

	  colors[i].pixel=i;
	  colors[i].flags=DoRed|DoBlue|DoGreen;
	}
  XStoreColors(display,cmap,colors,palette_size);
  XFlush(display);
}
  
void graph_show()
{
  extern char *bufferdata;
  XImage *im;
  char *tdat,*line,*p,*q;
  int x,y,i,yd;
  int xscale,yscale;

  xscale=dis_x/win_x;yscale=dis_y/win_y;

  if(!(tdat=malloc(dis_x*dis_y)))panic("unable to allocate for XImage!");

  if(!(line=malloc(dis_x)))panic("unable to allocate for scanline");

  /* for each line in the actual window */

  for(yd=y=0;y<win_y;y++)
	{
	  p=bufferdata+win_x*y;	/* get ptr to scanline */

	  /* copy the data into the line, scaling up */
	  for(q=line,x=0;x<win_x;x++)
		{
		  for(i=0;i<xscale;i++)*q++ = *p;
		  p++;
		}

	  /* now copy the line into the tmp buffer n times */
	  for(i=0;i<yscale;i++)
		memcpy(tdat+dis_x*yd++,line,dis_x);
	}

  /* copy the buffer to the screen */

  im=XCreateImage(display,visual,8,ZPixmap,0,
				  tdat,dis_x,dis_y,8,dis_x);

  XPutImage(display,window,gc,im,0,0,0,0,dis_x,dis_y);
  XDestroyImage(im);
  free(line);
  buffer_clear();
}

void graph_shutdown()
{
  if(display)XCloseDisplay(display);
}

